//+------------------------------------------------------------------+
//|                                                        modem.mqh |
//|                                         Copyright 2013, Decanium |
//|                                          http://www.decanium.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2013, Decanium"
#property link      "http://www.decanium.com"

#include <ComPort.mqh>

//---    !!!    -  ( ) - .
#define fECHO_ON              0x000001//"ATE1"
//: ( )
#define fGET_MANUFACTURER     0x000002//"AT+CGMI"
//: SIEMENS
#define fGET_DEVICE           0x000004//"AT+CGMM"
//: M55
#define fGET_SIM_STATUS       0x000008//"AT^SCKS?"
//: ^SCKS: 0,1 (x,0 -  ; x,1 -  ;   ,  sysinfo)
#define fGET_SYSINFO          0x000010//"AT^SYSINFO"
//: ^SYSINFO:2,1,0,3,1,,3 ( 1  ",," -  -; 1 - , 0 -    -, 2-4 -   USIM)
#define fGET_NET_REGISTRATION 0x000020//"AT+CREG?"
//: +CREG: 0,1 (x,0 -  ; x,1 - ; x,2 -  ; x,3 -  ; x,4 - ; x,5 -  )
#define fGET_OP_NAME          0x000040//"AT+COPS?"
//: +COPS: 0,0,"MTS UKR",0 (  "MTS UKR"   25501)
#define fSW_TO_PDU            0x000080//"AT+CMGF=0"
//: ( ;       )
#define fSW_DISP_CALLERID     0x000100//"AT+CLIP=1"
//: ( )
#define fGET_MODEM_STATUS     0x000200//AT+CPAS
//: +CPAS: 0 (0 - Ready, 3 - Incoming call (ring), 4 - active call)
#define fGET_RSSI             0x000400//"AT+CSQ"
//: +CSQ: 23,99 (  -  rssi;      Ready)
#define fGET_USSD_SUPPORT     0x000800//"AT+CUSD?"
//: +CUSD: 0
#define fSW_SILENT_MODE       0x001000//"AT+CALM=1"
//: ( ;   .   )
#define fGET_BAT_STAT	      0x002000//"AT+CBC"
//: +CBC: 0,90	(  -  ,  -    )
#define fGET_BALANCE          0x004000//  *101#
//:   
#define fGET_FREE_SMS         0x008000//  *101*4#
//:    
#define fGET_SMSC_NUMBER      0x010000//"AT+CSCA?"     
//: +CSCA: "+38050000501",145
#define fGET_NEW_MESSAGE      0x020000//"AT+CMGL=0"    
//:   
#define fGET_SMS_MEMSIZE      0x040000//"AT+CPMS="SM","SM","SM""     -
//: +CPMS: 10,15,10,15,10,15 ( 10 - -   , 15 -   )

#define fMASK_ALL          0x7FFFF//
#define fMASK_DEV_PRESENT  0x00007//
#define fMASK_NET_REG      0x607E0//     
#define fMASK_PER_REFR 		0x02620//     

#define SMS_LIST_SIZE      16//  
#include <sms.mqh>
//+------------------------------------------------------------------+
//| USSD_STR                                                         |
//+------------------------------------------------------------------+
struct USSD_STR
  {
   bool              enable;               //   ussd-
   int               tout;                 //     
   string            response;             //  USSD 
   bool              busy;                 // ()
   int               cur_flag;             //  ( )
   int               refr_bal_period_cnt;  //     
   int               refr_bal_period_init;
   int               refr_sms_period_cnt;  //     
   int               refr_sms_period_init;
  };
//+------------------------------------------------------------------+
//| MODEM_STR                                                        |
//+------------------------------------------------------------------+
struct MODEM_STR
  {
   bool              init_ok;      //  

   string            manufacturer; //
   string            device;       //
   int               sim_stat;     // SIM
   int               net_reg;      //   
   int               status;       // 
   string            op;           //
   int               rssi;         // 
   string            sms_sca;      // -
   int               bat_stat;     //  (, , ,  ,  ..., . -1)
   int               bat_charge;   //    (   bat_stat)

   double            bal;          //   (. -10000)
   string            exp_date;     //     
   int               sms_free;     //  
   int               sms_free_cnt; //   

   int               sms_mem_size; //  
   int               sms_mem_used; //   

   string            incoming;     //  
  };
//+------------------------------------------------------------------+
//| MODEM_DRV_STR                                                    |
//+------------------------------------------------------------------+
struct MODEM_DRV_STR
  {
   uint              flags_init_ok; // "  "
   uint              flags_no_supp; // "  "

   string            sms_out_msg;   //   

   USSD_STR          ussd;
   int               refr_period_cnt;

   bool              new_incoming_call;
   bool              new_incoming_sms;
   bool              sms_memory_full;
   bool              chng_state;
  };

#define AT_TRANSMIT_TOUT   20  // x50ms = 1s
#define AT_RESPONSE_TOUT   20  // x50ms = 1s
#define AT_MAX_ATTEMPTS     3  //  -     

#define AT_STRINGS         64  //        
//+------------------------------------------------------------------+
//| AT_CMD_STR                                                       |
//+------------------------------------------------------------------+
struct AT_CMD_STR
  {
   int               res;//  : -1   , 0 -  ERROR, 1 -  OK 
   int               resp_tout;
   uchar             txbuf[COM_TX_BUFSIZE];
   int               txn;
   int               txnn;//   (  )
   int               rxn;
   bool              busy;
   //

   string            s[AT_STRINGS];
   int               sn;
  };

MODEM_DRV_STR mdrv;
MODEM_STR modem;
AT_CMD_STR at;
/////
//+------------------------------------------------------------------+
//| rssi_to_str                                                      |
//+------------------------------------------------------------------+
string rssi_to_str(int rssi)
  {
// 0 -113 dBm or less
// 1 -111 dBm
// 2-30 -109 to -53 dBm
// 31 -51dBm or more
// 99 Unknown
   string str;
//
   double r;
   if(rssi>=0 && rssi<=30)
     {
      r=(rssi-2)*2-109;
      str=DoubleToString(r,0)+" dBm";
     }
   else
      str="n/a";
//   
   return(str);
  }
//+------------------------------------------------------------------+
//| hex_to_str                                                       |
//+------------------------------------------------------------------+
string hex_to_str(uchar h)
  {
   string hexbuf[16]={"0","1","2","3","4","5","6","7","8","9","A","B","C","D","E","F"};
   string s="";
   s+= hexbuf[(h>>4)&0x0F];
   s+= hexbuf[h&0x0F];
   return(s);
  }
//+------------------------------------------------------------------+
//| str_to_hex                                                       |
//+------------------------------------------------------------------+
uchar str_to_hex(string str)
  {
   uchar buf[2];
   StringToCharArray(str,buf,0,2,CP_ACP);
   uchar h=0;
   if(buf[0]>=48 && buf[0]<=57)
      h=buf[0]-48;
   else if(buf[0]>=65 && buf[0]<=70)
                   h=buf[0]-55;
   h=h<<4;
   if(buf[1]>=48 && buf[1]<=57)
      h+=buf[1]-48;
   else if(buf[1]>=65 && buf[1]<=70)
                   h+=buf[1]-55;
   return(h);
  }

uchar pdutab[8][3]=
  {
     {0x7F,0x00,0x00},{0x3F,0x01,0x07},
     {0x1f,0x03,0x06},{0x0f,0x07,0x05},
     {0x07,0x0F,0x04},{0x03,0x1F,0x03},
     {0x01,0x3F,0x02},{0x00,0x7F,0x01}
  };
//+------------------------------------------------------------------+
//| decodePDU                                                        |
//+------------------------------------------------------------------+
string decodePDU(string str)
  {
   uchar outbuf[161];
   string s="";
   int len = StringLen(str);
   uchar val;
   outbuf[0]=0;
   int i=0,ii=0;
//
   while(ii<(len/2))
     {
      val=str_to_hex(StringSubstr(str,ii*2,2));
      //
      outbuf[i]|=uchar((val&pdutab[i&0x07][0])<<(i&0x07));
      i++;
      outbuf[i]=(val>>pdutab[i&0x07][2])&pdutab[i&0x07][1];
      if((i&0x07)==0x07)
        {
         i++;
         outbuf[i]=0;
        }
      //   
      ii++;
     }
   return(CharArrayToString(outbuf,0,i,CP_ACP));
  }
//+------------------------------------------------------------------+
//| encodePDU                                                        |
//+------------------------------------------------------------------+
string encodePDU(string str,bool tussd)
  {
   uchar inbuf[];
   int len=StringToCharArray(str,inbuf,0,-1,CP_ACP)-1;
   if(len>160)
      len=160;
   int ii=0;
   uchar out=uchar((inbuf[ii]>>(ii&0x07))&pdutab[ii&0x07][0]);
   string s="";
   ii++;
//
   while(ii<len)
     {
      out|=(pdutab[ii&0x07][1]&inbuf[ii])<<pdutab[ii&0x07][2];
      if((ii&0x07)!=0x00)
         s+=hex_to_str(out);
      out=uchar((inbuf[ii]>>(ii&0x07))&pdutab[ii&0x07][0]);
      ii++;
     }
   if((ii&0x07)!=0x00)
      s+=hex_to_str(out);
//
   if(tussd==true)
     {// USSD-   (  )   0x1A
      int l=StringLen(s);
      if(StringSubstr(s,l-2,2)=="00")
         s=StringSubstr(s,0,l-2)+"1A";
     }
//
   return(s);
  }
//+------------------------------------------------------------------+
//| SendUSSD                                                         |
//+------------------------------------------------------------------+
bool SendUSSD(string str)
  {
   if(mdrv.ussd.enable==true)
     {
      if(mdrv.ussd.busy==true)
         return(false);
      if(SendATcommand("AT+CUSD=1,"+encodePDU(str,true)+",15",at)==true)
        {
         mdrv.ussd.tout=inp_ussd_request_tout;
         mdrv.ussd.busy=true;
         return(true);
        }
      else
         return(false);
     }
   return(false);
  }
//+------------------------------------------------------------------+
//| read_str                                                         |
//+------------------------------------------------------------------+
bool read_str(string &str,bool echo)
  {
//---    (  \r\n  \r\n)    -
//---   false,          
//---        0d0a ... 0d0a
//--- !!!     0d0a    0d !!!
//--- !!!       0d0a,    !!!
//---  echo==true,   "" 0d0a   "0a"
   int txn,rxn;
//--- ,       -
   TrComPortGetQueue(PortID,rxn,txn);
   if(rxn>0)
     {
      //Print("New Rx data: "+DoubleToString(rxn,0)+"bytes");
      if(rxn>COM_RX_BUFSIZE)
         rxn=COM_RX_BUFSIZE;
      TrComPortReadArray(PortID,rx_buf,rxn,-1);
      rx_cnt+=rxn;
      rx_str+=CharArrayToString(rx_buf,0,rxn,CP_ACP);
     }
//---      
   int len=StringLen(rx_str);
   if(len<1)
      return(false);
//---    0x0d   
   int pos=StringFind(rx_str,"\r",0);
   if(pos==-1)
      return(false);
//---  0x0d    
   if(pos>0 && echo==true)
     {
      str=StringSubstr(rx_str,0,pos);
      if(len>(pos+1))
         rx_str=StringSubstr(rx_str,pos+1,len-(pos+1));
      else
         rx_str="";
      return(true);
     }
//---
   if(StringSubstr(rx_str,pos,2)=="\r\n")
     {
      pos+=2;
      int pos2=StringFind(rx_str,"\r\n",pos);
      if(pos2==-1)
        {//--- ,     
         pos2=StringFind(rx_str,"> ",pos);
         if(pos2==-1)
            return(false);
         str=StringSubstr(rx_str,pos,(pos2+2)-pos);
        }
      else
         str=StringSubstr(rx_str,pos,pos2-pos);
      if(len>(pos2+2))
         rx_str=StringSubstr(rx_str,pos2+2,len-(pos2+2));
      else
         rx_str="";
      return(true);
     }
//      
   return(false);
  }
//+------------------------------------------------------------------+
//| SendAT_data                                                      |
//+------------------------------------------------------------------+
bool SendAT_data(AT_CMD_STR &atc)
  {
   int result=-1;
   int rxn,txn;
//
   atc.resp_tout=AT_TRANSMIT_TOUT;
   while(atc.resp_tout>0)
     {
      TrComPortGetQueue(PortID,rxn,txn);
      if(txn==0)
        {
         result=TrComPortWriteArray(PortID,atc.txbuf,atc.txn,-1);
         if(result<0)
            tx_err++;//Print("   ,  "+DoubleToString(result,0));
         else
           {
            tx_cnt+=atc.txn;
            break;
           }
        }
      Sleep(50);//ms
      atc.resp_tout--;
     }
//
   if(result<0)
      return(false);
   else
      return(true);
  }
//+------------------------------------------------------------------+
//| SendATcommand                                                    |
//+------------------------------------------------------------------+
bool SendATcommand(string str,AT_CMD_STR &atc)
  {
   atc.busy=true;
//Print("  "+str);
   int pos1;//, pos2, res;
            //int si = 0;//   
   int stat;// 0 -  , 1 -  ,  ,  , 2 -  , 3 -  ERROR
   string strtmp;
   int len=StringLen(str);
   StringToCharArray(str,atc.txbuf,0,len,CP_ACP);
   atc.txbuf[len]=0x0d;
//atc.txbuf[len+1]= 0x0a;
   atc.txn=len+1;//len+2;
                 //
   if(SendAT_data(atc)==false)
     {
      Print("      ");
      atc.busy=false;
      return(false);
     }
//
   int tout=AT_RESPONSE_TOUT;
   atc.resp_tout=tout;
//---  
//---    ,  ,  - 
   strtmp="";
   pos1=0;
   stat=0;
   int attempts=0;
//
   atc.txnn=0;
   atc.sn=0;//-   
   atc.s[0]="";
   int si=0;//   
   while(atc.resp_tout>0)
     {
      Sleep(50);//ms
      if(read_str(strtmp,true)==true)
        {//--- - 
         //Print("- : "+strtmp);
         if(newstr_proc(strtmp)==false)
           {//--- ,    
            //Print("- : "+strtmp);
            //
            if(stat==0)
              {//---  
               //Print("  \""+str+"\"");
               if(StringFind(strtmp,str,0)!=-1)
                 {
                  //Print(" ");
                  stat=1;
                 }
              }
            else if(stat==1)
              {
               if(StringFind(strtmp,"OK",0)!=-1)
                 {
                  //Print(" OK");
                  stat=2;
                  atc.sn = si;
                  atc.res=1;
                  break;
                 }
               else if(StringFind(strtmp,"ERROR",0)!=-1)
                 {
                  //Print(" ERROR");
                  stat=3;
                  atc.sn = si;
                  atc.res=0;
                  break;
                 }
               else if(StringFind(strtmp,"> ",0)!=-1)
                 {//---   
                  //---      PDU     
                  //Print("  ");
                  str = mdrv.sms_out_msg;
                  len = StringLen(str);
                  StringToCharArray(mdrv.sms_out_msg,atc.txbuf,0,len,CP_ACP);
                  atc.txbuf[len]=0x1a;
                  atc.txn=len+1;
                  if(SendAT_data(atc)==false)
                    {
                     Print("      ");
                     atc.busy=false;
                     return(false);
                    }
                  tout=AT_RESPONSE_TOUT*20;
                  atc.resp_tout=tout;//---      ,   
                  stat=1;//---  OK
                 }
               else
                 {
                  //Print(" ( "+DoubleToString(si,0)+"): "+strtmp);
                  if(si<AT_STRINGS)
                    {
                     atc.s[si]=strtmp;
                     StringTrimLeft(atc.s[si]);
                    }
                  si++;
                  if(si<AT_STRINGS)
                     atc.s[si]="";
                  //---     
                  atc.resp_tout=tout;
                 }
              }
           }
        }
      atc.resp_tout--;
      if(atc.resp_tout==0 && stat==0)
        {//---  ,    
         if(attempts<AT_MAX_ATTEMPTS)
           {//---    
            //Print("<<<<<<< RETRANSMIT >>>>>>>");
            if(SendAT_data(atc)==false)
              {
               Print("      ");
               atc.busy=false;
               return(false);
              }
            atc.resp_tout=tout;
            attempts++;
           }
         //else
         //   Print("<<<<<<< TIMEOUT >>>>>>>");   
        }
     }
   atc.busy=false;
   if(stat<2 || stat==4)
      return(false);
//   
   return(true);
  }
//+------------------------------------------------------------------+
//| AT_PARAM_STR                                                     |
//+------------------------------------------------------------------+
struct AT_PARAM_STR
  {
   int               val;
   string            str;
  };
//+------------------------------------------------------------------+
//| ATparam_is_string                                                |
//+------------------------------------------------------------------+
bool ATparam_is_string(string &str_in,string &str_out)
  {
   int pos= StringFind(str_in,"\"",0);
   if(pos!=-1)
     {
      int pos2= StringFind(str_in,"\"",pos+1);
      if(pos2!=-1)
        {
         if((pos2-pos)>1)
            str_out=StringSubstr(str_in,pos+1,pos2-(pos+1));
         else
            str_out="";
         return(true);
        }
     }
   return(false);
  }
//+------------------------------------------------------------------+
//| GetATparam(string &str,int index,AT_PARAM_STR &par)              |
//| str -                                                      |
//| sep -                                                 |
//| index -                                           |
//| par - ,                    |
//|      true                        |
//+------------------------------------------------------------------+
bool GetATparam(string &str,int index,AT_PARAM_STR &par)
  {
   string ss[];
   ushort u_sep=StringGetCharacter(",",0);
//
   int k=StringSplit(str,u_sep,ss);
   if(k<1)
      return(false);
   if(index>=k)
      return(false);
   if(index==0)
     {
      string s=ss[0];
      u_sep=StringGetCharacter(":",0);
      k=StringSplit(s,u_sep,ss);
      if(k>1)
        {
         if(ATparam_is_string(ss[1],par.str)==false)
            par.val=(int)StringToInteger(ss[1]);
        }
      else
        {
         if(ATparam_is_string(s,par.str)==false)
            par.val=(int)StringToInteger(s);
        }
     }
   else
     {
      if(ATparam_is_string(ss[index],par.str)==false)
         par.val=(int)StringToInteger(ss[index]);
     }
   return(true);
  }
//+------------------------------------------------------------------+
//| CheckFlag                                                        |
//+------------------------------------------------------------------+
int CheckFlag(int flag)
  {
   if((mdrv.flags_init_ok&flag)==flag)
      return(1);
   else if((mdrv.flags_no_supp&flag)==flag)
      return(-1);
   else
      return(0);
  }
//+------------------------------------------------------------------+
//| QueryModem                                                       |
//+------------------------------------------------------------------+
bool QueryModem()
  {
   AT_PARAM_STR par;
//---
   if((mdrv.flags_init_ok|mdrv.flags_no_supp)==fMASK_ALL)
      return(true);//---   
//---
   if(CheckFlag(fECHO_ON)==0)
     {//---  
      if(SendATcommand("ATE1",at)==true)
        {
         if(at.res==1)
            mdrv.flags_init_ok|=fECHO_ON;
        }
      else
        {
         SendESC();
         return(false);
        }
     }
   mdrv.chng_state=true;
//---
   if(CheckFlag(fGET_MANUFACTURER)==0)
     {//---   
      if(SendATcommand("AT+CGMI",at)==true)
        {
         if(at.res==1)
           {
            modem.manufacturer=at.s[0];
            mdrv.flags_init_ok|=fGET_MANUFACTURER;
           }
        }
      else
         return(false);
     }
   mdrv.chng_state=true;
//---
   if(CheckFlag(fGET_DEVICE)==0)
     {//---   
      if(SendATcommand("AT+CGMM",at)==true)
        {
         if(at.res==1)
           {
            modem.device=at.s[0];
            mdrv.flags_init_ok|=fGET_DEVICE;
           }
        }
      else
         return(false);
     }
   mdrv.chng_state=true;
//---
   if(CheckFlag(fGET_SIM_STATUS)==0)
     {//---   -
      if(SendATcommand("AT^SCKS?",at)==true)
        {
         if(at.res==1)
           {//---  
            par.val=-1;
            if(GetATparam(at.s[0],1,par)==true)
              {
               modem.sim_stat=par.val;
               //---  - 
               mdrv.flags_init_ok|=fGET_SIM_STATUS;
               //--- ,  fGET_SYSINFO   
               mdrv.flags_no_supp|=fGET_SYSINFO;
              }
           }
         else
           {//---   
            mdrv.flags_no_supp|=fGET_SIM_STATUS;
           }
        }
      else
         return(false);
     }
   mdrv.chng_state=true;
//
   if(CheckFlag(fGET_SYSINFO)==0)
     {//---    (      -)  
      if(SendATcommand("AT^SYSINFO",at)==true)
        {
         if(at.res==1)
           {//---  
            par.val=-1;
            if(GetATparam(at.s[0],4,par)==true)
              {
               if(par.val!=1)
                  modem.sim_stat=0;
               else
                  modem.sim_stat=1;
               mdrv.flags_init_ok|=fGET_SYSINFO;
              }
           }
         else if(at.res==0)
           {//---   
            mdrv.flags_no_supp|=fGET_SYSINFO;
           }
        }
      else
         return(false);
     }
   mdrv.chng_state=true;
//
   if(CheckFlag(fGET_NET_REGISTRATION)==0 && modem.sim_stat==1)
     {//---      (   -)
      if(SendATcommand("AT+CREG?",at)==true)
        {
         if(at.res==1)
           {
            par.val=-1;
            if(GetATparam(at.s[0],1,par)==true)
              {
               modem.net_reg=par.val;
               mdrv.flags_init_ok|=fGET_NET_REGISTRATION;
              }
           }
        }
      else
         return(false);
     }
   mdrv.chng_state=true;
//
   if(CheckFlag(fGET_OP_NAME)==0 && (modem.net_reg==1 || modem.net_reg==5))
     {//---   ,       (    )
      if(SendATcommand("AT+COPS?",at)==true)
        {
         if(at.res==1)
           {
            par.str="n/a";
            if(GetATparam(at.s[0],2,par)==true)
              {
               //Print("fGET_OP_NAME: "+par.str+", "+DoubleToString(par.val,0));
               modem.op=par.str;
               mdrv.flags_init_ok|=fGET_OP_NAME;
              }
           }
        }
      else
         return(false);
     }
   mdrv.chng_state=true;
//
   if(CheckFlag(fSW_TO_PDU)==0 && (modem.net_reg==1 || modem.net_reg==5))
     {//---    PDU (    )
      if(SendATcommand("AT+CMGF=0",at)==true)
        {
         if(at.res==1)
            mdrv.flags_init_ok|=fSW_TO_PDU;
        }
      else
         return(false);
     }
   mdrv.chng_state=true;
//
   if(CheckFlag(fSW_DISP_CALLERID)==0 && (modem.net_reg==1 || modem.net_reg==5))
     {//---    
      if(SendATcommand("AT+CLIP=1",at)==true)
        {
         if(at.res==1)
            mdrv.flags_init_ok|=fSW_DISP_CALLERID;
        }
      else
         return(false);
     }
   mdrv.chng_state=true;
//
   if(CheckFlag(fGET_MODEM_STATUS)==0 && (modem.net_reg==1 || modem.net_reg==5))
     {//---   
      if(SendATcommand("AT+CPAS",at)==true)
        {
         if(at.res==1)
           {
            par.val=-1;
            if(GetATparam(at.s[0],0,par)==true)
              {
               modem.status=par.val;
               mdrv.flags_init_ok|=fGET_MODEM_STATUS;
              }
           }
        }
      else
         return(false);
     }
   mdrv.chng_state=true;
//
   if(inp_sms_service_center!="")
     {//---   -  
      modem.sms_sca=inp_sms_service_center;
      mdrv.flags_no_supp|=fGET_SMSC_NUMBER;
     }
   mdrv.chng_state=true;
//
   if(CheckFlag(fGET_SMSC_NUMBER)==0 && modem.status==0)
     {//---   - 
      if(SendATcommand("AT+CSCA?",at)==true)
        {
         if(at.res==1)
           {//---   - 
            par.str="";
            if(GetATparam(at.s[0],0,par)==true)
              {
               modem.sms_sca=par.str;
               //Print(" -: "+modem.sms_sca);
               mdrv.flags_init_ok|=fGET_SMSC_NUMBER;
              }
            else
               Print("   -");
           }
         else
           {//--- 
            mdrv.flags_no_supp|=fGET_SMSC_NUMBER;
           }
        }
      else
         return(false);
     }
   mdrv.chng_state=true;
//   
   if(CheckFlag(fGET_RSSI)==0 && modem.status==0)
     {//---      (    )
      if(SendATcommand("AT+CSQ",at)==true)
        {
         if(at.res==1)
           {
            par.val=-1;
            if(GetATparam(at.s[0],0,par)==true)
              {
               modem.rssi=par.val;
               mdrv.flags_init_ok|=fGET_RSSI;
               modem.init_ok=true;
              }
           }
        }
      else
         return(false);
     }
   mdrv.chng_state=true;
//
   if(CheckFlag(fGET_USSD_SUPPORT)==0)
     {//---     USSD-
      if(SendATcommand("AT+CUSD?",at)==true)
        {
         if(at.res==1)
           {//--- USSD- 
            mdrv.ussd.enable=true;
            mdrv.flags_init_ok|=fGET_USSD_SUPPORT;
           }
         else
           {//--- USSD-  
            mdrv.ussd.enable=false;
            mdrv.flags_no_supp|=fGET_USSD_SUPPORT;
            mdrv.flags_no_supp|=fGET_BALANCE;
            mdrv.flags_no_supp|=fGET_FREE_SMS;
           }
        }
      else
         return(false);
     }
   mdrv.chng_state=true;
//---
   if(CheckFlag(fSW_SILENT_MODE)==0)
     {//---    (  )
      if(SendATcommand("AT+CALM=1",at)==true)
        {
         if(at.res==1)
           {//---   
            mdrv.flags_init_ok|=fSW_SILENT_MODE;
           }
         else
           {//---    
            mdrv.flags_no_supp|=fSW_SILENT_MODE;
           }
        }
      else
         return(false);
     }
   mdrv.chng_state=true;
//---
   if(CheckFlag(fGET_BAT_STAT)==0)
     {//---   
      if(SendATcommand("AT+CBC",at)==true)
        {
         if(at.res==1)
           {//
            par.val=-1;
            if(GetATparam(at.s[0],0,par)==true)
              {
               modem.bat_stat=par.val;
               if(GetATparam(at.s[0],1,par)==true)
                  modem.bat_charge= par.val;
               mdrv.flags_init_ok|=fGET_BAT_STAT;
              }
           }
         else
           {//---   
            mdrv.flags_no_supp|=fGET_BAT_STAT;
           }
        }
      else
         return(false);
     }
   mdrv.chng_state=true;
//
   if(CheckFlag(fGET_SMS_MEMSIZE)==0)
     {//---        ( / -)
      //--- if(SendATcommand("AT+CPMS=\"SM\",\"SM\",\"SM\"", at)==true)
      if(SendATcommand("AT+CPMS=\"MT\",\"MT\",\"MT\"",at)==true)
        {
         if(at.res==1)
           {//
            par.val=-1;
            if(GetATparam(at.s[0],0,par)==true)
              {
               if(par.val>=0)
                 {
                  modem.sms_mem_used=par.val;
                  par.val=-1;
                  if(GetATparam(at.s[0],1,par)==true)
                    {
                     if(par.val>=0)
                       {
                        modem.sms_mem_size=par.val;
                        if(modem.sms_mem_used==modem.sms_mem_size)
                           mdrv.sms_memory_full=true;
                        mdrv.flags_init_ok|=fGET_SMS_MEMSIZE;
                       }
                    }
                 }
              }
           }
        }
      else
         return(false);
     }
   mdrv.chng_state=true;
//---
   if(CheckFlag(fGET_BALANCE)==0)
     {
      if(inp_ussd_get_balance!="")
        {
         if(SendUSSD(inp_ussd_get_balance)==true)
            mdrv.ussd.cur_flag=fGET_BALANCE;// 
         //---     ,       flags_init_ok
        }
      else
         mdrv.flags_no_supp|=fGET_BALANCE;
     }
   mdrv.chng_state=true;
//---
   if(CheckFlag(fGET_FREE_SMS)==0)
     {

      if(inp_ussd_get_sms_cnt!="")
        {
         if(SendUSSD(inp_ussd_get_sms_cnt)==true)
            mdrv.ussd.cur_flag=fGET_FREE_SMS;// 
         //---     ,       flags_init_ok
        }
      else

         mdrv.flags_no_supp|=fGET_FREE_SMS;
     }
   mdrv.chng_state=true;
   if(CheckFlag(fGET_NEW_MESSAGE)==0)
     {//---   ,     
      if(SendATcommand("AT+CMGL=0",at)==true)//  "AT+CMGL=1"
        {
         int sms_index=-1; //   
         int smsi=0;       //   smsl.list[]
         datetime prev_scts=0;
         for(int j=0; j<AT_STRINGS; j++)
           {
            if(at.s[j]=="")
               break;
            else
              {
               if(StringFind(at.s[j],"+CMGL",0)!=-1)
                 {
                  par.val=-1;
                  if(GetATparam(at.s[j],0,par)==true)
                    {
                     if(par.val>=0)
                       {
                        if(par.val<modem.sms_mem_size)
                           sms_index=par.val;
                       }
                    }
                 }
               else
                 {
                  if(sms_index>=0 && smsi<SMS_LIST_SIZE)
                    {
                     if(ParseSMS(at.s[j],smsl.list[smsi])==true)
                       {
                        smsl.list[smsi].index=sms_index;//    (  )
                        Print("<"+DoubleToString(sms_index,0)+">"+smsl.list[smsi].text);
                        if(smsl.list[smsi].scts.time>=prev_scts)
                          {// smsl.last     - 
                           smsl.last = smsi;
                           prev_scts = smsl.list[smsi].scts.time;
                          }
                        smsi++;
                        mdrv.new_incoming_sms=true;
                       }
                     else
                        Print("<<ERROR>>");
                     sms_index=-1;
                    }
                 }
              }
           }
         mdrv.flags_init_ok|=fGET_NEW_MESSAGE;
        }
     }
   mdrv.chng_state=true;
//
   return(true);
  }
//+------------------------------------------------------------------+
//| DelSMSbyIndex                                                    |
//+------------------------------------------------------------------+
bool DelSMSbyIndex(int index)
  {
   if(SendATcommand("AT+CMGD="+DoubleToString(index,0),at)==true)
     {
      mdrv.sms_memory_full=false;
      return(true);
     }
   else
      return(false);
  }
//+------------------------------------------------------------------+
//| ReadSMSbyIndex                                                   |
//+------------------------------------------------------------------+
bool ReadSMSbyIndex(int index,INCOMING_SMS_STR &sms)
  {
   if(SendATcommand("AT+CMGR="+DoubleToString(index,0),at)==true)
     {
      if(StringFind(at.s[0],"+CMGR",0)!=-1)
        {
         if(ParseSMS(at.s[1],sms)==true)
            return(true);
        }
     }
//   
   return(false);
  }
//+------------------------------------------------------------------+
//| SendESC                                                          |
//+------------------------------------------------------------------+
void SendESC()
  {
   at.txbuf[0]= 0x1a;
   at.txbuf[1]= 0x0d;
   at.txbuf[2]= 0x0a;
   at.txn=3;
   TrComPortWriteArray(PortID,at.txbuf,at.txn,-1);
   tx_cnt+=at.txn;
  }
//+------------------------------------------------------------------+
//| InitModem                                                        |
//+------------------------------------------------------------------+
void InitModem()
  {
//---
   mdrv.flags_init_ok=0;
   mdrv.flags_no_supp=0;
//---
   modem.manufacturer="n/a";
   modem.device="n/a";
   modem.sim_stat=-1;
   modem.net_reg=-1;
   modem.op="n/a";
   modem.status=-1;
   modem.rssi=-1;
   mdrv.ussd.enable=false;
   mdrv.ussd.tout=0;
   mdrv.ussd.busy=false;
   mdrv.ussd.refr_bal_period_cnt=0;
   mdrv.ussd.refr_bal_period_init=3600*inp_refr_bal_period;
   mdrv.ussd.refr_sms_period_cnt=0;
   mdrv.ussd.refr_sms_period_init=3600*inp_refr_smscnt_period;
   modem.bat_stat=-1;
   modem.bat_charge=0;
   modem.bal=-10000;
   modem.sms_free_cnt=-1;
   modem.sms_free=-1;
   modem.exp_date="n/a";
   modem.incoming="n/a";
//--- modem.call_action="n/a";
   mdrv.new_incoming_call=false;
   mdrv.new_incoming_sms=false;
   smsl.last=-1;
   mdrv.sms_memory_full=false;
   modem.sms_mem_size=0;
//---
   modem.init_ok=false;
   at.busy=false;
   mdrv.chng_state=false;
//---
   QueryModem();
  }
//+------------------------------------------------------------------+
//| ModemTimerProc                                                   |
//|                                   |
//+------------------------------------------------------------------+
void ModemTimerProc()
  {
   string s;
//---
   for(int i=0;i<10;i++)
     {
      if(read_str(s,false)==true)
         newstr_proc(s);
      else
         break;
     }
//---
   if(modem.sim_stat==0)
     {
      mdrv.flags_init_ok = mdrv.flags_init_ok&(~(fGET_SIM_STATUS|fGET_SYSINFO|fGET_SMSC_NUMBER));
      mdrv.flags_no_supp = mdrv.flags_no_supp&(~(fGET_SIM_STATUS|fGET_SYSINFO|fGET_SMSC_NUMBER));
     }
//---
   if(++mdrv.refr_period_cnt>=inp_refr_period)
     {
      mdrv.flags_init_ok=mdrv.flags_init_ok&(~(fGET_BAT_STAT|
                                             fGET_NET_REGISTRATION|
                                             fGET_MODEM_STATUS|
                                             fGET_RSSI|
                                             fGET_SMS_MEMSIZE|
                                             fGET_NEW_MESSAGE));
      mdrv.refr_period_cnt=0;
     }
//---
   if(++mdrv.ussd.refr_bal_period_cnt>=mdrv.ussd.refr_bal_period_init)
     {//---     
      mdrv.flags_init_ok=mdrv.flags_init_ok&(~(fGET_BALANCE));
      mdrv.ussd.refr_bal_period_cnt=0;
     }
//---
   if(++mdrv.ussd.refr_sms_period_cnt>=mdrv.ussd.refr_sms_period_init)
     {//---      
      mdrv.flags_init_ok=mdrv.flags_init_ok&(~(fGET_FREE_SMS));
      mdrv.ussd.refr_sms_period_cnt=0;
     }
   QueryModem();
//---
   if(mdrv.chng_state==true)
     {//---   
      ModemChState();
      mdrv.chng_state=false;
     }
//
   if(mdrv.sms_memory_full==true)
      SMSMemoryFull(modem.sms_mem_used);
//
   if(modem.sms_free_cnt>=0)
      modem.sms_free=inp_free_sms_daily-modem.sms_free_cnt;
   else
      modem.sms_free=-1;
//
   if(mdrv.ussd.tout>0)
     {
      if(--mdrv.ussd.tout==0)
        {
         Print("   USSD-.");
         mdrv.ussd.busy=false;
        }
     }
//---
   if(mdrv.new_incoming_sms==true)
     {
      IncomingSMS(smsl.list[smsl.last]);
      mdrv.new_incoming_sms=false;
     }
//---
   if(mdrv.new_incoming_call==true)
     {
      if(SendATcommand("AT+CHUP",at)==true)// 
         IncomingCall(modem.incoming);
      mdrv.new_incoming_call=false;
     }
  }
//+------------------------------------------------------------------+
//| get_val_right(double &v,string &str,int pos0,int pos1)           |
//|                                   |
//| v -                                           |
//| str -                                                      |
//| pos0 -                                        |
//| pos1 -                                       |
//|  false,                              |
//+------------------------------------------------------------------+
bool get_val_right(double &v,string &str,int pos0,int pos1)
  {
   string digs[12]={"0","1","2","3","4","5","6","7","8","9",".","-"};
   if(pos1<pos0)
      return(false);

   int st=0;
   string s="";
   for(int i=pos1;i>=pos0;i--)
     {
      string ss=StringSubstr(str,i,1);
      bool ok=false;
      for(int j=0;j<12;j++)
        {
         if(ss==digs[j])
           {
            ok=true;
            break;
           }
        }
      if(ok==true)
        {
         if(st==0)
            st=1;
         s=ss+s;
        }
      else
        {
         if(st==1)
            break;
        }
     }

   if(s=="")
      return(false);
   v=StringToDouble(s);
   return(true);
  }
//+------------------------------------------------------------------+
//| bool get_valstr_left(string &vs,string &str,int pos0,int pos1)   |
//|                         |
//| v -                                           |
//| str -                                                      |
//| pos0 -                                        |
//| pos1 -                                       |
//|  false,                              |
//+------------------------------------------------------------------+
bool get_valstr_left(string &vs,string &str,int pos0,int pos1)
  {
   string digs[12]={"0","1","2","3","4","5","6","7","8","9",".","-"};
   if(pos1<pos0)
      return(false);

   int st=0;
   string s="";
   for(int i=pos0;i<=pos1;i++)
     {
      string ss=StringSubstr(str,i,1);
      bool ok=false;
      for(int j=0;j<12;j++)
        {
         if(ss==digs[j])
           {
            ok=true;
            break;
           }
        }
      if(ok==true)
        {
         if(st==0)
            st=1;
         s=s+ss;
        }
      else
        {
         if(st==1)
            break;
        }
     }

   if(s=="")
      return(false);
   vs=s;
//v = StringToDouble(s);
   return(true);
  }
//+------------------------------------------------------------------+
//| newstr_proc                                                      |
//|   ,                        |
//|  -   -  true                    |
//+------------------------------------------------------------------+
bool newstr_proc(string &s)
  {
   int res;
//
//Print(">"+s);
//---
   res=StringFind(s,"RSSI:",0);
   if(res!=-1)
     {
      string ss=StringSubstr(s,res+5,2);
      StringTrimRight(ss);
      modem.rssi=(int)StringToInteger(ss);
      return(true);
     }
//---
   res=StringFind(s,"+CLIP:",0);
   if(res!=-1)
     {
      int pos1 = StringFind(s, "\"", res+6);
      int pos2 = StringFind(s, "\"", pos1+1);
      if((pos2-pos1)<2)
         modem.incoming="hidden";
      else
         modem.incoming=StringSubstr(s,pos1+1,pos2-(pos1+1));
      //
      mdrv.new_incoming_call=true;
      //
      return(true);
     }
//---
   res=StringFind(s,"+CUSD:",0);
   if(res!=-1)
     {
      int pos1 = StringFind(s, "\"", res+6);
      int pos2 = StringFind(s, "\"", pos1+1);
      mdrv.ussd.response=decodePDU(StringSubstr(s,pos1+1,pos2-(pos1+1)));
      if(mdrv.ussd.cur_flag!=0)//    ussd-
         Print("USSD response: "+mdrv.ussd.response);
      //---
      if(mdrv.ussd.response!="")
        {
         if(mdrv.ussd.cur_flag==fGET_BALANCE)
           {//---      
            pos1=StringFind(mdrv.ussd.response,inp_ussd_bal_suffix,0);
            if(pos1!=-1)
              {//---   
               double v;
               if(get_val_right(v,mdrv.ussd.response,0,pos1-1)==true)
                  modem.bal=v;
              }
            //
            pos1=StringFind(mdrv.ussd.response,inp_ussd_exp_prefix,0);
            if(pos1!=-1)
              {//---      
               string vs;
               if(get_valstr_left(vs,mdrv.ussd.response,pos1+StringLen(inp_ussd_exp_prefix),StringLen(mdrv.ussd.response)-1)==true)
                  modem.exp_date=vs;
              }
           }
         //
         if(mdrv.ussd.cur_flag==fGET_FREE_SMS)
           {//---       
            pos1=StringFind(mdrv.ussd.response,inp_ussd_sms_suffix,0);
            if(pos1!=-1)
              {//---     
               double v;
               if(get_val_right(v,mdrv.ussd.response,0,pos1-1)==true)
                  modem.sms_free_cnt=(int)v;
              }
           }
        }
      if(mdrv.ussd.cur_flag!=0)
        {//---  - ,      flags_init_ok
         mdrv.flags_init_ok|=mdrv.ussd.cur_flag;
         mdrv.ussd.cur_flag=0;
        }
      mdrv.ussd.busy=false;
      mdrv.ussd.tout=0;
      return(true);
     }
//
   return(false);
  }
//+------------------------------------------------------------------+
